﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;

namespace RobowiiL
{
    public class Recognition
    {

        List<Model> ModelList;
        //int[] indexCount;
        double[] indexCount, accelIndexCount, gyroIndexCount;
        int K = 3;

        List<ModelNameCount> realModel = new List<ModelNameCount>();
        List<ModelNameCount> realAccelModel = new List<ModelNameCount>();
        List<ModelNameCount> realGyroModel = new List<ModelNameCount>();


        public Recognition(List<Model> ModelList)
        {
            this.ModelList = ModelList;
            indexCount = new double[ModelList.Count];

            accelIndexCount = new double[ModelList.Count];
            gyroIndexCount = new double[ModelList.Count];
        }

        public Recognition(List<Model> ModelList, int k)
        {
            this.ModelList = ModelList;
            indexCount = new double[ModelList.Count];

            accelIndexCount = new double[ModelList.Count];
            gyroIndexCount = new double[ModelList.Count];

            this.K = k;
        }

//*************************************** INIZIO KNN MIGLIORATO *************************************************************

        //dati in ingresso due matrici contenenti per ognuno dei 250 punti l'indice della gesture più vicina a quella ricreata
        //calcola il puntaggio di ogni gesture in base alla distanza.
        //Inserisce il risultato in accelIndexCount e gyroIndexCount
        private void FindClassModel(int[][] minAccelDistanceOfModels, int[][] minGyroDistanceOfModels)
        {
            int index = 0, indexModel = 0, indexKMin = 0;

            for (indexKMin = 0; indexKMin < K; indexKMin++)
            {
                for (index = 0; index < 250; index++)
                {
                    if (minAccelDistanceOfModels[indexKMin][index] != -1)
                    {
                        indexModel = minAccelDistanceOfModels[indexKMin][index];
                        accelIndexCount[indexModel] += K - indexKMin;
                    }

                    if (minGyroDistanceOfModels[indexKMin][index] != -1)
                    {
                        indexModel = minGyroDistanceOfModels[indexKMin][index];
                        gyroIndexCount[indexModel] += K - indexKMin;
                    }
                }
            }
        }


        double FindMin(double[][] array, int index)
        {
            int count = ModelList.Count;

            double min = array[0][index];

            for (int indexModel = 1; indexModel < count; indexModel++)
            {
                if (array[indexModel][index] < min)
                    min = array[indexModel][index];
            
            }

            return min;
        }

        double FindMax(double[][] array, int index)
        {
            int count = ModelList.Count;

            double max = array[0][index];

            for (int indexModel = 1; indexModel < count; indexModel++)
            {
                if (array[indexModel][index] > max)
                    max = array[indexModel][index];

            }

            return max;
        }

        private void InizializeDistanceModel(int[][] minAccelDistanceOfModels, int[][] minGyroDistanceOfModels)
        {
            for (int i = 0; i < K; i++)
                for (int j = 0; j < 250; j++)
                {
                    minAccelDistanceOfModels[i][j] = -1;
                    minGyroDistanceOfModels[i][j] = -1;
                }


        }


        //Crea due matrici che contengono gli indici dei 
        //k modelli più vicini per ogni punto
        private void FindMinimumValueForEveryPoint(double[][] accelDistanceOfModels, double[][] gyroDistanceOfModels)
        {
            int index = 0, indexModel = 0, indexKMin, count = this.ModelList.Count;

            double minAccelDist, tempAccelDist;
            double minGyroDist, tempGyroDist;
            double maxGyroDist, maxAccelDist;

            int[][] minAccelDistanceOfModels = new int[K][];
            int[][] minGyroDistanceOfModels = new int[K][];

            for (indexKMin = 0; indexKMin < K; indexKMin++)
            {
                minAccelDistanceOfModels[indexKMin] = new int[250];
                minGyroDistanceOfModels[indexKMin] = new int[250];
            }

            InizializeDistanceModel(minAccelDistanceOfModels, minGyroDistanceOfModels);

            //per tutti i valori
            for (index = 0; index < 250; index++)
            {
                //per tutti gli elementi da inserire fino a K
                for (indexKMin = 0; indexKMin < K; indexKMin++)
                {
                    //prendo il minimo della colonna
                    //delle accelerazioni
                    minAccelDist = FindMin(accelDistanceOfModels, index);
                    //dei giroscopi
                    minGyroDist = FindMin(gyroDistanceOfModels, index);

                    //prendo il massimo sulla colonna
                    //delle accelerazioni
                    maxAccelDist = FindMax(accelDistanceOfModels, index);
                    //dei giroscopi
                    maxGyroDist = FindMax(gyroDistanceOfModels, index);

                    //di ogni modello
                    for (indexModel = 0; indexModel < this.ModelList.Count; indexModel++)
                    {

                        tempAccelDist = accelDistanceOfModels[indexModel][index];
                        tempGyroDist = gyroDistanceOfModels[indexModel][index];

                        //prendo il modello che ha per il punto index la distanza delle accelerazioni minore
                        if (tempAccelDist == minAccelDist)
                        {
                            accelDistanceOfModels[indexModel][index] = maxAccelDist;

                            minAccelDistanceOfModels[indexKMin][index] = indexModel;
                        }
                        //prendo il modello che ha per il punto index la distanza dei giroscopi minore
                        if (tempGyroDist == minGyroDist)
                        {
                            gyroDistanceOfModels[indexModel][index] = maxGyroDist;

                            minGyroDistanceOfModels[indexKMin][index] = indexModel;
                        }

                    }
                }
            }

            FindClassModel(minAccelDistanceOfModels, minGyroDistanceOfModels);
        }

        //dati in ingresso i modelli delle gesture
        //restituisce il nome della gesture riconosciuta
        public string KNNRecognitionMatrix(Model model)
        {
            double x, y, z, speedX, speedY, speedZ;
            int index = 0, indexModel = 0;
            double accelDistance = 0;
            double degreeDistance = 0;

            //creo due matrici con un numero di righe grande 
            //quanto il numero dei modelli inseriti
            double[][] accelDistanceOfModels = new double[this.ModelList.Count][];
            double[][] gyroDistanceOfModels = new double[this.ModelList.Count][];

            Model temp;

            //parso tutti i modelli e prendo il più vicino a punto in indice index
            for (indexModel = 0; indexModel < this.ModelList.Count; indexModel++)
            {
                temp = this.ModelList[indexModel];

                accelDistanceOfModels[indexModel] = new double[250];
                gyroDistanceOfModels[indexModel] = new double[250];

                for (index = 0; index < 250; index++)   //parso tutti i punti
                {
                    //for

                    x = model.GetX(index);
                    y = model.GetY(index);
                    z = model.GetZ(index);

                    speedX = model.GetXSpeed(index);
                    speedY = model.GetYSpeed(index);
                    speedZ = model.GetZSpeed(index);

                    accelDistance = CoordDistance(x, y, z, temp.GetX(index), temp.GetY(index), temp.GetZ(index));
                    degreeDistance = SpeedDistance(speedX, speedY, speedZ, temp.GetXSpeed(index), temp.GetYSpeed(index), temp.GetZSpeed(index));

                    accelDistanceOfModels[indexModel][index] = accelDistance;
                    gyroDistanceOfModels[indexModel][index] = degreeDistance;
                }
            }

            //trovo la gesture a minima distanza da quella inserita
            FindMinimumValueForEveryPoint(accelDistanceOfModels, gyroDistanceOfModels);           

            //genero i due array, uno per gli accelerometri
            //e uno per i giroscopi con l'indice dei K modelli che hanno
            //ottenuto il punteggio più alto
            int[] KNNAccel = ThreeMax(accelIndexCount);
            int[] KNNGyro = ThreeMax(gyroIndexCount);

            //trovo il modello che più si avvicina alla gesture e restituisco tale valore
            //al chiamante (RoboMain).
            return ModelRecognition(KNNAccel, KNNGyro);
        }

        //trova i k modelli che risultano più vicini alla gesture eseguita.
        private int[] ThreeMax(double[] count)
        {
            int[] KNNArray = new int[K];
            int index = 0;


            for (int kIndex = 0; kIndex < K; kIndex++)
            {
                for (index = 0; index < count.Length; index++)
                {
                    if (count[index] == count.Max())
                    {
                        KNNArray[kIndex] = index;
                        count[index] = 0;
                        break;
                    }
                }
            }
            return KNNArray;
        }



  /*      private int[] InitializeRealModelIndex(int[] accelIndex, int[] gyroIndex)
        {
            int[] realIndex = new int[2*K];

            for (int i = 0; i < accelIndex.Count(); i++)
                realIndex[i] = accelIndex[i];

            for (int i = accelIndex.Count(); i < gyroIndex.Count(); i++)
                realIndex[i] = gyroIndex[i];

            return realIndex;
        }
        */
        //date in ingresso le matrici con i K modelli con il punteggio più alto
        //restituisco il modello più vicino alla gesture eseguita.
        private string ModelRecognition(int[] accelIndex, int[] gyroIndex)
        {
           
            realAccelModel.Add(new ModelNameCount(this.ModelList[accelIndex[0]].GetName()));
            realGyroModel.Add(new ModelNameCount(this.ModelList[gyroIndex[0]].GetName()));

            realAccelModel[0].IncrementCount();
            realGyroModel[0].IncrementCount();
           
            for (int j = 1; j < K; j++)
            {
                EqualModel(realAccelModel, accelIndex, j);
                EqualModel(realGyroModel, gyroIndex, j);
            }


            //trovo il numero massimo che indica quanti modelli dello stesso tipo sono presenti nell'array accelIndex
            int maxCountAccelModel = 0;
            foreach (ModelNameCount m in realAccelModel)
            {
                if (m.GetCount() > maxCountAccelModel)
                    maxCountAccelModel = m.GetCount();

            }
            
                
            //trovo il numero massimo che indica quanti modelli dello stesso tipo sono presenti nell'array gyroIndex
            int maxCountGyroModel = 0;
            foreach (ModelNameCount m in realGyroModel)
            {
                if (m.GetCount() > maxCountGyroModel)
                    maxCountGyroModel = m.GetCount();

            }

            string gyroModel = null;

            //Per ogni modello controllo se corrisponde al massimo trovato tramite i punteggi
            //Se trovo un modello presente nell'array degli accelerometri dello stesso tipo
            //di quello più comune nei giroscopi
            //Restituisco il modello in questione.
            for (int j = 0; j < realGyroModel.Count ; j++)
            {
                int count = realGyroModel[j].GetCount();
                for (int i = 0; i < realAccelModel.Count; i++)
                {
                    if (count == maxCountGyroModel && realAccelModel[i].GetName().ToString() == realGyroModel[j].GetName().ToString())
                    {
                        gyroModel = realGyroModel[j].GetName();
                        break;
                    }
                }
            }

            if (gyroModel != null)
                return gyroModel;
            


            //Nel caso in cui non sia soddisfatta la condizione precedente eseguo la stessa ricerca
            //ma utilizzando al posto dei giroscopi gli accelerometri e viceversa.
            for (int j = 0; j < realAccelModel.Count; j++)
            {
                int count = realAccelModel[j].GetCount();
                for (int i = 0; i < realGyroModel.Count; i++)
                {
                    if (count == maxCountAccelModel && realGyroModel[i].GetName().ToString() == realAccelModel[j].GetName().ToString())
                    {
                        return realAccelModel[j].GetName();
                    
                    }
                } 
            }
            
            //se neppure questa condizione è soddisfatta non è stato possibile riconoscere la gesture.
            return "Recognition Failed";
        }

        //Data:
        //- la lista dei modelli;
        //- l'array con i K modelli con il punteggio maggiore;
        //- l'indice attuale nell'array con i K modelli più vicini
        //Conta quanti modelli sono stati trovati dello stesso tipo
        private void EqualModel(List<ModelNameCount> real, int[] index, int j)
        {
            bool find = false;

            for (int i = 0; i < real.Count; i++)
            {
                find = false;

                if (real[i].GetName().ToString() == this.ModelList[index[j]].GetName().ToString())
                {
                    real[i].IncrementCount();
                    find = true;
                    break;
                } 
            }

            if (!find)
            {
                real.Add(new ModelNameCount(this.ModelList[index[j]].GetName()));
                real[real.Count - 1].IncrementCount();
            }
        }


//*************************************** FINE PARTE DA TESTARE *************************************************************




        //Da riguardare assoulutamente
        public string KNNRecognition(Model model)
        {
            int index = 0, indexModel = 0, indexMinD = 0;
            double distance = 0, min = 0;
            double x, y, z, speedX, speedY, speedZ;
            bool roll, pitch, yaw;
            Model temp;
            // int countGyroOfMin, countGyroTemp;
            double minSpeedDistance, speedDistance;
            //indexCount = new int[ModelList.Count];

            for (index = 0; index < 250; index++)   //parso tutti i punti
            {
                //for

                x = model.GetX(index);
                y = model.GetY(index);
                z = model.GetZ(index);

                roll = model.GetRoll(index);
                pitch = model.GetPitch(index);
                yaw = model.GetYaw(index);

                speedX = model.GetXSpeed(index);
                speedY = model.GetYSpeed(index);
                speedZ = model.GetZSpeed(index);

                temp = this.ModelList[0];

                min = CoordDistance(x, y, z, temp.GetX(index), temp.GetY(index), temp.GetZ(index));
                //countGyroOfMin = GyroscopeDistance(roll, pitch, yaw, temp.GetRoll(index), temp.GetPitch(index), temp.GetYaw(index));
                minSpeedDistance = SpeedDistance(speedX, speedY, speedZ, temp.GetXSpeed(index), temp.GetYSpeed(index), temp.GetZSpeed(index));

                for (indexModel = 1; indexModel < this.ModelList.Count; indexModel++)//parso tutti i modelli e prendo il più vicino a punto in indice index
                {
                    //distanza
                    temp = this.ModelList[indexModel];
                    distance = CoordDistance(x, y, z, temp.GetX(index), temp.GetY(index), temp.GetZ(index));
                    //countGyroTemp = GyroscopeDistance(roll, pitch, yaw, temp.GetRoll(index), temp.GetPitch(index), temp.GetYaw(index));
                    speedDistance = SpeedDistance(speedX, speedY, speedZ, temp.GetXSpeed(index), temp.GetYSpeed(index), temp.GetZSpeed(index));


                    if (distance < min && speedDistance < minSpeedDistance)
                    {
                        indexMinD = indexModel;
                        //countGyroOfMin = countGyroTemp;
                        minSpeedDistance = speedDistance;
                        min = distance;
                    }

                }
                indexCount[indexMinD]++;
                //index count varibility 0->10
                /*   if (distance == 0)
                       indexCount[indexMinD] += 10;
                   else
                   {
                       double inc = 1 * (1 /distance);
                       if(inc >10)
                           inc = 9;
                       indexCount[indexMinD] += inc;//incremento il numero di punti appertenenti al modello di indice indexMinD
                    
                   }*/
            }
            //max array count
            //int K = 3;
            int[] KNNArray = new int[K];
            for (int i = 0; i < K; i++)
            {

                for (index = 0; index < indexCount.Length; index++)
                {
                    if (indexCount[index] == indexCount.Max())
                    {
                        KNNArray[i] = index;
                        indexCount[index] = 0;
                        break;
                    }
                }

            }
            string str = LastModelRecognitionProva(KNNArray);

            return str;
            //str è il comando da inviare (o mettere nella coda...) all'NXT
        }




        public string LastModelRecognitionProva(int[] index)
        {
            bool find = false;
            // int realModelCount = 0;

            realModel.Add(new ModelNameCount(this.ModelList[index[0]].GetName()));
            realModel[0].IncrementCount();
            //realModel[0].SetName(this.ModelList[index[0]].GetName());
            //realModelCount++;

            for (int j = 1; j < index.Length; j++)
            {
                for (int i = 0; i < realModel.Count; i++)
                {
                    find = false;
                    if (realModel[i].GetName().ToString() == this.ModelList[index[j]].GetName().ToString())
                    {
                        realModel[i].IncrementCount();
                        find = true;
                        break;
                    }
                }
                if (!find)
                {
                    realModel.Add(new ModelNameCount(this.ModelList[index[j]].GetName()));
                    realModel[realModel.Count - 1].IncrementCount();
                    //realModel[realModel.Count].SetName(this.ModelList[index[j]].GetName());
                    //realModelCount++;
                }
            }

            int max = 0;

            foreach (ModelNameCount m in realModel)
            {
                if (m.GetCount() > max)
                    max = m.GetCount();

            }


            foreach (ModelNameCount m in realModel)
            {
                if (m.GetCount() == max)
                    return m.GetName();
            }
            return "";
        }


        //****************** Distance *************************

        private double CoordDistance(double xModel, double yModel, double zModel, double xLModel, double yLModel, double zLModel)
        {
            return Math.Sqrt(((xModel - xLModel) * (xModel - xLModel) + (yModel - yLModel) * (yModel - yLModel) + (zModel - zLModel) * (zModel - zLModel)));
            // return 0;
        }//Distance

        private int GyroscopeDistance(bool rollModel, bool pitchModel, bool yawModel, bool rollLModel, bool pitchLModel, bool yawLModel)
        {
            int count = 0;

            if (rollModel == rollLModel)
                count++;
            if (pitchModel == pitchLModel)
                count++;
            if (yawModel == yawLModel)
                count++;

            return count;
        }

        private double SpeedDistance(double speedXModel, double speedYModel, double speedZModel, double speedXLModel, double speedYLModel, double speedZLModel)
        {
            return Math.Sqrt(((speedXModel - speedXModel) * (speedXModel - speedXModel) + (speedYModel - speedYLModel) * (speedYModel - speedYLModel) + (speedZModel - speedZLModel) * (speedZModel - speedZLModel)));
            // return 0;
        }//Distance

    }
}
