﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.IO.Ports;
using System.Threading;
using WiimoteLib;//from wiilib
using System.Collections;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;

namespace RobowiiL
{
    public partial class RoboMain : Form
    {
        //Delegate
        delegate void writeOnLogDelegate(string text);
        delegate void WriteCoordsInModelDelegate(double x, double y, double z, bool roll, bool pitch, bool yaw, double speedX, double speedY, double speedZ);
        delegate void CreateNewModelDelegate();
        delegate void WinMessageDelegate();
        
        //Attributes
        CreateNewModelDelegate cnm;
        writeOnLogDelegate wold;
        WriteCoordsInModelDelegate wcimd;
        WinMessageDelegate wmd;
        public WiimoteCollection mWC;
        private nxtCommand nxtC;
        private Wiimote mWiimote;
        private SerialPort BluetoothConnection  = new SerialPort();//predatore
        private SerialPort BluetoothConnection2 = new SerialPort();//preda
        public WiiControl wiiC;

        private int levelSelected;
        //ArrayList arrModel;
        private List<Model> ModelList = new List<Model>();
        private Model model;
        private int modelIndex = 0;

       // private FileStream file;
        private string fileName;
        private Recognition rec;
        public bool training = false;

        private bool allConnectionOK = false;//if false -> no connection to wiimote or nxt
                                             //if true -> two nxt and wiimote connected


        private bool game = false;//if false -> not playing
                                  //if true -> playing

        //timer & sound
        private System.Media.SoundPlayer myPlayer = new System.Media.SoundPlayer();
        private System.Windows.Forms.Timer timerLose = new System.Windows.Forms.Timer();
        
        //Level param
        private int secCounter = 7*60;
        private int numColpi = 1;
        
        //need ModelNameCount in order to find the real model in k-model array
        //List<ModelNameCount> realModel = new List<ModelNameCount>();

        //********************



        public RoboMain()
        {
            InitializeComponent();
            InitializeGestureType();
            wold = new writeOnLogDelegate(writeOnLog);
            wcimd = new WriteCoordsInModelDelegate(WriteCoordsInModel);
            cnm = new CreateNewModelDelegate(CreateNewModel);
            wmd = new WinMessageDelegate(WinMessage);
            bool fileOk = this.OpenFile("default.ges");//Open file di default
            if (!fileOk)
            {
                this.training = true;
                this.checkTraining.Checked = true;//spunta il checkbox del training
            }
        }


        //forse non serve più: IncrementModelIndex
        public void IncrementModelIndex()
        {
            this.modelIndex++;
        }



        private void InitializeGestureType()
        {
        
        this.ComboModel.Items.Add("Avanti");
        this.ComboModel.Items.Add("Indietro");
        this.ComboModel.Items.Add("Destra");
        this.ComboModel.Items.Add("Sinistra");
        this.ComboModel.Items.Add("Shake");
        this.ComboModel.Items.Add("Quadrato");
        this.ComboModel.Items.Add("Cerchio");
        this.ComboModel.Text = this.ComboModel.Items[0].ToString();

        }



        //LastModelRecognition
        //riconosce il modello a cui appartiene l'ultima serie di dati acquisiti
        //
        public void LastModelRecognition()
        {
            this.rec = new Recognition(this.ModelList);

            //this.model è l'ultimo dato acquisito
            string realModel = rec.KNNRecognition(this.model);
            string realModelTest = rec.KNNRecognitionMatrix(this.model);
            
            if (realModel != null)
            {
                this.externwriteOnLog(realModel);
                this.externwriteOnLog("Test:"+realModelTest);
                //mandare alla coda.....

                command currCmd = new command(realModelTest, 0.5);
                if (this.nxtC != null)
                {
                    this.nxtC.ParseCmd(currCmd);
                }
            }
            else
                this.externwriteOnLog("Gesture non riconosciuta");
        }//LastModelRecognition


        public TextBox getTextBox()
        {
            return this.Log;
        }

        public void externwriteOnLog(string str)
        {
            DateTime CurrTime = DateTime.Now;
            this.Log.Invoke(wold, new object[] { CurrTime.TimeOfDay +": "+ str });
        }

        private void writeOnLog(string str)
        {

            if (this.Log.Lines.Length < 100)
            {
                this.Log.Text += str;
                this.Log.Text += Environment.NewLine;
            }
            else
            {
                this.Log.Text = str;
            }
            this.Log.SelectionStart = this.Log.Text.Length;
            this.Log.ScrollToCaret();
        }

        //public void WriteCoordsInModel(double x, double y, double z, bool roll, bool pitch, bool yaw)
        public void WriteCoordsInModel(double x, double y, double z, bool roll, bool pitch, bool yaw, double speedX, double speedY, double speedZ)
        {
        //    this.ModelList[this.modelIndex].AddValue(x, y, z, roll, pitch, yaw, speedX, speedY, speedZ);
            this.ModelList[this.ModelList.Count-1].AddValue(x, y, z, roll, pitch, yaw, speedX, speedY, speedZ);
        }

        public void WriteCoordsInRecognitionModel(double x, double y, double z, bool roll, bool pitch, bool yaw, double speedX, double speedY, double speedZ)
        {

            this.model.AddValue(x, y, z, roll, pitch, yaw, speedX, speedY, speedZ);
        }

        public void CreateNewRecognitionModel()
        {
            this.model = new Model();
        
        }

        public void ExternCreateNewModel()
        {
            this.ComboModel.Invoke(cnm);
            
        }

        private void CreateNewModel()
        {
            // old call: this.ModelList.Add(new Model(this.ClassifierName.Text.ToString()));
            this.ModelList.Add(new Model(this.ComboModel.Text.ToString()));
            this.GestureNameTB.Text = this.ComboModel.Text.ToString();
        }//CreateNewModel()

        private int connectWiiMote()
        {
            mWC = new WiimoteCollection();
            int index = 0;

            try
            {
                mWC.FindAllWiimotes();
            }
            catch (WiimoteNotFoundException ex)
            {
                MessageBox.Show(ex.Message, "Wiimote not found error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            catch (WiimoteException ex)
            {
                MessageBox.Show(ex.Message, "Wiimote error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Unknown error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            foreach (Wiimote wm in mWC)
            {
                // create a new tab
                //TabPage tp = new TabPage("Wiimote " + index);
                //tabWiimotes.TabPages.Add(tp);

                // create a new user control
                //WiimoteInfo wi = new WiimoteInfo(wm);
                mWiimote = wm;
                //tp.Controls.Add(wi);

                // setup the map from this wiimote's ID to that control
                //mWiimoteMap[wm.ID] = wi;

                // connect it and set it up as always
               
                //wm.WiimoteExtensionChanged += wm_WiimoteExtensionChanged;
                try
                {
                    wm.Connect();
                }
                catch(WiimoteException ex) 
                {
                    MessageBox.Show(ex.Message, "Wiimote not found error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return 0;
                }
                if (wm.WiimoteState.ExtensionType != ExtensionType.BalanceBoard)
                    wm.SetReportType(InputReport.IRExtensionAccel, IRSensitivity.Maximum, true);
                //GIGIGIGI
                wm.WiimoteChanged += wiiC.wm_WiimoteChanged;

                wm.SetLEDs(index+1);
                index += 1;
            }
            return index;
        }


       


      /*  private void BtnAddNew_Click(object sender, EventArgs e)
        {
            if (this.ClassifierName.Text.ToString() == "")
            {
                MessageBox.Show("Ops!", "Insert first classifier name", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            this.ComboModel.Items.Add(this.ClassifierName.Text.ToString()); 
            
            //aggiungi al pop up 
        }*/

        private void RoboMain_Load(object sender, EventArgs e)
        {

        }

        private void Save_Click(object sender, EventArgs e)
        {
            //creo un oggetto da serializzare
            ObjectToSerialize objectToSerialize = new ObjectToSerialize();
            //setto la lista di modelli come contenuto dell'oggetto che verrà
            //scritto sul disco
            objectToSerialize.Models = this.ModelList;

            //questo è l'oggetto che si occupa effettivamente della scrittura
            Serializer serializer = new Serializer();
            serializer.SerializeObject(this.fileName, objectToSerialize);
        }//Save_Click


        private void SaveAs_Click(object sender, EventArgs e)
        {
            this.SaveAsDialog.Filter = "ges files (*.ges)|*.ges";
            this.SaveAsDialog.ShowDialog();
            this.fileName = this.SaveAsDialog.FileName;
            
            this.Save_Click(this, null);

        }//SaveAs_Click



        private void SaveAsDialog_FileOk(object sender, CancelEventArgs e)
        {

        }

        private bool OpenFile(string PFileName)
        {
            this.fileName = PFileName;//set curr filename
            

            //create serialize object
            ObjectToSerialize objectToSerialize = new ObjectToSerialize();
            //create serializer object
            Serializer serializer = new Serializer();

            //deserialize
            objectToSerialize = serializer.DeSerializeObject(this.fileName);

            //if file isn't new file or empty:
            if (objectToSerialize != null)
            {
                this.ModelList = objectToSerialize.Models;//init modelList
                this.Forward.Enabled = true;
                this.Back.Enabled = true;

                this.Invalidate();

                return true;
            }
            this.ModelList = new List<Model>();

            this.Forward.Enabled = false;
            this.Back.Enabled = false;
            this.Invalidate();

            return false;
            
        }//OpenFile()

        private void Open_Click(object sender, EventArgs e)
        {
            //Cartella di partenza
            //this.OpenFileDialog.InitialDirectory = "D:\\";

            //configura l'estensione di default
            this.OpenFileDialog.Filter = "ges files (*.ges)|*.ges|All files (*.*)|*.*";
            this.OpenFileDialog.FilterIndex = 1;

            //Configura il nome di default
            this.OpenFileDialog.FileName = "";

            this.OpenFileDialog.ShowDialog();//show open dialog box

            //Get il filename e path selezionato
            this.OpenFile(this.OpenFileDialog.FileName);
        }//Open_Click


        private void AboutBtn_Click(object sender, EventArgs e)
        {
            this.AboutDialog.ShowDialog();
        }//AboutBtn_Click

        public void InterpolateLastModel()
        {
            //this.ModelList[this.modelIndex].Interpolate();
            this.ModelList[this.ModelList.Count-1].Interpolate();
        
        }

        public void InterpolateRecognitionModel()
        {
            this.model.Interpolate();
        }

        public void PrintLastModel()
        {
            this.modelIndex = this.ModelList.Count - 1;
            this.Invalidate();
        }
       
        public void PrintModel()
        {
            this.Invalidate();
        }

        private void Menu_ItemClicked(object sender, ToolStripItemClickedEventArgs e)
        {

        }


        //********************Connection click events
        private void InitializeMotionPlus()
        {
            if (mWiimote != null)
            {
                this.writeOnLog("Motion Plus Initializing");
                mWiimote.InitializeMotionPlus();
                this.writeOnLog("Motion Plus Initialized");
            }
            else {
                this.writeOnLog("No Wiimote Connected"); 
            }
        
        }


        private void ConnectWii_Click(object sender, EventArgs e)
        {

            this.ConnectWii.Enabled = false;
            if (mWiimote != null)
            {
                mWiimote.Disconnect();
                this.ConnectWii.Text = "ConnectWii";
                mWiimote = null;
            }
            else
            {
                wiiC = new WiiControl(this);
                int connectedWiiMoteNum = connectWiiMote();
                //if found and connect to any wiimote
                if (connectedWiiMoteNum > 0)
                {
                    this.ConnectWii.Text = "Disconnect";
                    this.writeOnLog("Wii Connect\n");
                    this.InitializeMotionPlus();

                }
                else
                {
                    MessageBox.Show("No Wiimote Connected", "No Wiimote Connected", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    this.writeOnLog("No wiimote connected\n");
                }

            }
            this.ConnectWii.Enabled = true;
        }



        private void ConnectAll_Click(object sender, EventArgs e)
        {
            this.ConnectAll.Enabled = false;
            
            //Se si è già connessi si è premuto il tasto di disconnessione
            //chiudi connessione
            if (BluetoothConnection.IsOpen)
            {
                //chiusura delle connessioni a NXT
                BluetoothConnection.Close();
                this.ConnectAll.Text = "ConnectAll";
                
                //Chiusura connessioni WiiMote
                if (mWiimote != null)
                {
                    mWiimote.Disconnect();
                    mWiimote = null;
                }
            }
            else//Ci si vuole connettere a Wiimote e NXT
            {
                try//recupero nomi porte com
                {
                    this.BluetoothConnection.PortName   = this.portName.Text.Trim();//predatore
                    this.BluetoothConnection2.PortName  = this.portName2.Text.Trim();//preda
                }
                catch (ArgumentException ex)
                {
                    MessageBox.Show(ex.Message, "No port is selected", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    this.ConnectAll.Enabled = true;
                    return;
                }

                try//Apro connessioni verso NXT
                {
                    BluetoothConnection.Open();
                    BluetoothConnection2.Open();
                }
                catch(IOException ex) 
                {
                    MessageBox.Show(ex.Message, "Wrong port selected", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    this.ConnectAll.Enabled = true;
                    return;
                }

                BluetoothConnection.ReadTimeout = 1500;
                BluetoothConnection2.ReadTimeout = 1500;

                nxtC = new nxtCommand(BluetoothConnection, BluetoothConnection2, this);
                //Fine apertura connessione NXT

                //Apertura connessione wiiMote
                wiiC = new WiiControl(this, nxtC);
                int connectedWiiMoteNum = connectWiiMote();
                //if found and connect to any wiimote
                if (connectedWiiMoteNum > 0)
                {
                    this.writeOnLog("Wii Connected\n");
                }
                else
                {
                    MessageBox.Show("NoWiimote Connected", "NoWiimote Connected", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    this.writeOnLog("No wiimote connected\n");
                    return;
                }

                this.InitializeMotionPlus();

                //Connessione avvenuta con successo
                this.ConnectAll.Text = "Disconnect";
                this.writeOnLog("NXT Connected\n");
                allConnectionOK = true;
            }
            //Connessioni non attive
            this.ConnectAll.Enabled = true;

        }//ConnectAll_Click

        private void TimerEnd(Object myObject, EventArgs myEventArgs)
        {
            //controllo se il tempo è scaduto
            //            if (this.secCounter < (5*60))
            if (this.secCounter > (0))
            {
                //se non è scaduto incremento
                this.secCounter -= 1;
                //e setto il nuovo tempo
                int minTemp = (int)(this.secCounter/60);
                int secTemp = this.secCounter - (minTemp*60);
                string temp;
                
                if(secTemp >9)
                    temp = minTemp.ToString() + ":" + secTemp.ToString();
                else
                    temp = minTemp.ToString() + ":0" + secTemp.ToString();

                this.Timer.Text = temp;
                return;
            }

            this.TimerStop();

            this.LoseMessage();

        }//TimerEnd

        public void TimerStop()
        {
            this.timerLose.Stop();
        }

        private void TimerReset()
        {
            if(this.levelSelected < 6)
                this.secCounter = 7 * 60;
            else
                this.secCounter = 5 * 60;
            
            //to test
            //this.secCounter =  5;
        }//TimerReset


        private void trainingToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.training = !this.training;
            this.checkTraining.Checked = this.training;
            this.writeOnLog("Training\n" + this.training.ToString());
        }

        private void Back_Click(object sender, EventArgs e)
        {
            if (this.modelIndex>0)
                this.modelIndex--;
            this.GestureNameTB.Text = this.ModelList[this.modelIndex].GetName().ToString();
            this.PrintModel();
        }

        private void Forward_Click(object sender, EventArgs e)
        {
            if (this.modelIndex == this.ModelList.Count)
                this.modelIndex--;

            else if (this.modelIndex < this.ModelList.Count-1)
                this.modelIndex++;
            this.GestureNameTB.Text = this.ModelList[this.modelIndex].GetName().ToString();
            this.PrintModel();
        }

        private void checkTraining_Click(object sender, EventArgs e)
        {
            this.training = !this.training;
            this.writeOnLog("Training\n" + this.training.ToString());
        }

        private void BTN_GAME_START_Click(object sender, EventArgs e)
        {

            //controllo la connettività
            //se non c'è connettività message box
            //se c'è connettività trasforma in stop game e fai partire il timer
            if (!allConnectionOK)
            {
                MessageBox.Show("You must first connect WiiMote & NXT", "Error connection", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            if (!game)//se il gioco non è attivo -> viene attibvato
            {
                BTN_GAME_START.Text = "STOP";
                game = !game;
                //Inizia il gioco, Inizia il timer
                //Funzione che gestisce l'evento timer
                this.timerLose.Tick += new EventHandler(TimerEnd);

                levelSelected = GetLevelSelected();
                if (levelSelected == 0)
                    this.writeOnLog("ERROR");
                else
                {
                    this.writeOnLog("Selected Level : " + levelSelected);
                    this.SetLevelParam();
                    this.nxtC.sendLevel(levelSelected);
                 
                }
                //Set timer
                this.TimerReset();
    
                // Sets the timer interval to 1 seconds.
                this.timerLose.Interval = 1000;

                this.timerLose.Start();
            }
            else//stoppo e resetto il gioco
            {
                StopAndReset();  
            }
        }

        private void StopAndReset()
        {
            BTN_GAME_START.Text = "Start game!";
            this.timerLose.Stop();
            this.TimerReset();
            game = !game;
            //fermo la preda
            this.nxtC.sendLevel(9);
            //è possibile riselezionare il livello

            this.timerLose.Tick -= TimerEnd;
        }


        int GetLevelSelected()
        {
            if (radioButton1.Checked)
                return 1;
            if (radioButton2.Checked)
                return 2;
            if (radioButton3.Checked)
                return 3;
            if (radioButton4.Checked)
                return 4;
            if (radioButton5.Checked)
                return 5;
            if (radioButton6.Checked)
                return 6;

            return -1;
        }

        private void SetLevelParam()
        {
            if (this.levelSelected > 3)
                numColpi = 2;
            else
                numColpi = 1;
        }//SetLevelParam


        public int Hit()
        {
            //restituisce il numero di colpi mancanti per uccidere la preda
            this.numColpi--;
            return this.numColpi;
        }//Hit

        public bool Alive()
        {
            //restituisce false se la preda è morta
            //true se è viva
            if (numColpi <= 0)
                return false;
            return true;
        }//Alive

        public void LoseMessage()
        {

            this.nxtC.sendLevel(9);
            myPlayer.SoundLocation = @"lose.wav";
            myPlayer.Play();
            
            MessageBox.Show("YOU LOSE!!!", "YOU LOSE!!!", MessageBoxButtons.OK);
           
            // Restarts the timer and increments the counter.
            //this.timerLose.Enabled = true;???? MA NON FUNZIONA!!!
            StopAndReset();
            
        }

        public void ExternWinMessage()
        {
            this.BTN_GAME_START.Invoke(wmd);
        }

        public void WinMessage()
        {

            this.nxtC.sendLevel(9);

            myPlayer.SoundLocation = @"win.wav";
            myPlayer.Play();

            StopAndReset();
            MessageBox.Show("YOU WIN!!!", "YOU WIN!!!", MessageBoxButtons.OK);

  
        }

        private void AboutDialog_Load(object sender, EventArgs e)
        {

        }
    }
}
